
class InvalidPlacement(Exception):
    pass

class ExistingBrick(InvalidPlacement):
    pass

class OutsideBounds(InvalidPlacement):
    pass

class InvalidBrick(InvalidPlacement):
    pass

class Step:
    def __init__(self, x_max, y_max):
        # https://stackoverflow.com/questions/10668341/create-3d-array-using-python
        self.grid = [["◖◗" for x in range(x_max)] for y in range(y_max)]
        self.x_max = x_max
        self.y_max = y_max

    def place_brick(self, x_corner, y_corner, x_size, y_size):
        # Check bounds
        if (x_corner < 0):
            raise OutsideBounds("x_corner must be at least 0")
        if (y_corner < 0):
            raise OutsideBounds("y_corner must be at least 0")
        if (x_corner + x_size > self.x_max):
            raise OutsideBounds(f"placing a {x_size} x {y_size} brick starting at {x_corner}, {y_corner} exceeds maximum x dimension")
        if (y_corner + y_size > self.y_max):
            raise OutsideBounds(f"placing a {x_size} x {y_size} brick starting at {x_corner}, {y_corner} exceeds maximum y dimension")
        
        if (x_size <= 0):
            raise InvalidBrick("x_size must be at least 1")
        if (y_size <= 0):
            raise InvalidBrick("y_size must be at least 1")

        # Check for overlap
        for x in range(x_corner, x_corner+ x_size):
            for y in range(y_corner, y_corner + y_size):
                if (self.grid[y][x] != "◖◗"):
                    raise ExistingBrick("overlaps with existing brick at " + str(x) + ", " + str(y))

        # Corners
        if (x_size == 1 and y_size == 1):
            self.grid[y_corner][x_corner] = "██"
        elif (x_size == 1):
            self.grid[y_corner][x_corner] = "▙▟" #chr(0x1fb7c) + chr(0x1fb7f)
            self.grid[y_corner + y_size - 1][x_corner] = "▛▜" #chr(0x1fb7d) + chr(0x1fb7e)
        elif (y_size == 1):
            self.grid[y_corner][x_corner] = "██" #█" + chr(0x1fb80)
            self.grid[y_corner][x_corner + x_size - 1] = "██" #chr(0x1fb80) + "█"
        else:
            self.grid[y_corner][x_corner] = "▙▄" #chr(0x1fb7c) + "▁"
            self.grid[y_corner][x_corner + x_size - 1] = "▄▟" #"▁" + chr(0x1fb7f)
            self.grid[y_corner + y_size - 1][x_corner] = "▛▀" #chr(0x1fb7d) + "▔"
            self.grid[y_corner + y_size - 1][x_corner + x_size - 1] = "▀▜" #"▔" + chr(0x1fb7e)

        # Top and bottom
        for x in range(x_corner + 1, x_corner + x_size - 1):
            if (y_size == 1):
                self.grid[y_corner][x] = chr(0x1fb80) * 2
            else:
                self.grid[y_corner][x] = "▄▄" #"▁▁"
                self.grid[y_corner + y_size - 1][x] = "▀▀" #"▔▔"

        # Left and right
        for y in range(y_corner + 1, y_corner + y_size - 1):
            if (x_size == 1):
                self.grid[y][x_corner] = "▌▐" #"▏▕"
            else:
                self.grid[y][x_corner] = "▌ " #"▏ "
                self.grid[y][x_corner + x_size - 1] = " ▐" #" ▕"
            
    def display(self):
        for y in range(self.y_max-1, -1, -1):
            row = "".join(self.grid[y])
            if (y % 2 == 0):
                row = f"{y:>2}" + row
            else:
                row = "  " + row
            print(row)
        print("  " + "".join([f"{col:<2}  " for col in range(0, self.x_max, 2)]))


def create_step(x_max, y_max) -> Step:
    return Step(x_max, y_max)
